@@ -1,6 +1,8 @@
 #include "config.h"
 #include "xyzsh/xyzsh.h"
 
+#include <term.h>
+#include <termios.h>
 #include <string.h>
 #include <strings.h>
 #include <stdlib.h>
@@ -481,6 +483,11 @@ static sObject* access_object_compl(char* name, sObject** current)
     }
 }
 
+char* readline_filename_completion_null_generator(const char* a, int b)
+{
+    return NULL;
+}
+
 char** readline_on_complete(const char* text, int start, int end)
 {
     stack_start_stack();
@@ -528,17 +535,28 @@ char** readline_on_complete(const char* text, int start, int end)
         }
     }
 
+    sObject* readline = uobject_item(gRootObject, "rl");
+    if(readline && TYPE(readline) == T_UOBJECT) {
+        uobject_put(readline, "omit_head_of_completion_display_matches", STRING_NEW_GC("0", TRUE));
+    }
+
+    rl_completion_entry_function = readline_filename_completion_null_generator;
+
+/*
     if(SBLOCK(gReadlineBlock).mCompletionFlags & COMPLETION_FLAGS_TILDA) {
         char** result = rl_completion_matches(text, rl_username_completion_function);
         stack_end_stack();
         return result;
     }
-    else if(SBLOCK(gReadlineBlock).mStatmentsNum == 0) {
+    else 
+*/
+    if(SBLOCK(gReadlineBlock).mStatmentsNum == 0) {
         char** result = rl_completion_matches(text, all_program_completion);
         stack_end_stack();
         return result;
     }
-    else if(SBLOCK(gReadlineBlock).mCompletionFlags & COMPLETION_FLAGS_AFTER_REDIRECT) {
+    else if(SBLOCK(gReadlineBlock).mCompletionFlags & (COMPLETION_FLAGS_AFTER_REDIRECT|COMPLETION_FLAGS_AFTER_EQUAL)) {
+        rl_completion_entry_function = NULL;
         stack_end_stack();
         return NULL;
     }
@@ -551,6 +569,15 @@ char** readline_on_complete(const char* text, int start, int end)
             return result;
         }
         else {
+            if(statment->mCommandsNum == 1 && statment->mCommands[0].mArgsNum == 1) {
+                char* arg = statment->mCommands[0].mArgs[0];
+                if(strstr(arg, "/")) {
+                    rl_completion_entry_function = NULL;
+                    stack_end_stack();
+                    return NULL;
+                }
+            }
+
             sCommand* command = statment->mCommands + statment->mCommandsNum-1;
 
             /// get user completion ///
@@ -721,26 +748,55 @@ BOOL cmd_completion(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
         }
         /// output ///
         else if(runinfo->mBlocksNum == 0) {
-            int i;
-            for(i=1; i<runinfo->mArgsNumRuntime; i++) {
-                sObject* compl;
-                if(!get_object_from_str(&compl, runinfo->mArgsRuntime[i], runinfo->mCurrentObject, runinfo->mRunningObject, runinfo)) {
-                    return FALSE;
-                }
+            if(sRunInfo_option(runinfo, "-source")) {
+                int i;
+                for(i=1; i<runinfo->mArgsNumRuntime; i++) {
+                    sObject* compl;
+                    if(!get_object_from_str(&compl, runinfo->mArgsRuntime[i], runinfo->mCurrentObject, runinfo->mRunningObject, runinfo)) {
+                        return FALSE;
+                    }
+
+                    if(compl && TYPE(compl) == T_COMPLETION) {
+                        sObject* block = SCOMPLETION(compl).mBlock;
+                        if(!fd_write(nextout, SBLOCK(block).mSource, strlen(SBLOCK(block).mSource))) 
+                        {
+                            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine, runinfo->mArgs[0]);
+                            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
+                            return FALSE;
+                        }
+                        
+                    }
+                    else {
+                        err_msg("There is no object", runinfo->mSName, runinfo->mSLine, runinfo->mArgsRuntime[i]);
 
-                if(compl && TYPE(compl) == T_COMPLETION) {
-                    if(!run_object(compl, nextin, nextout, runinfo)) {
                         return FALSE;
                     }
                 }
-                else {
-                    err_msg("There is no object", runinfo->mSName, runinfo->mSLine, runinfo->mArgsRuntime[i]);
 
-                    return FALSE;
-                }
+                runinfo->mRCode = 0;
             }
+            else {
+                int i;
+                for(i=1; i<runinfo->mArgsNumRuntime; i++) {
+                    sObject* compl;
+                    if(!get_object_from_str(&compl, runinfo->mArgsRuntime[i], runinfo->mCurrentObject, runinfo->mRunningObject, runinfo)) {
+                        return FALSE;
+                    }
 
-            runinfo->mRCode = 0;
+                    if(compl && TYPE(compl) == T_COMPLETION) {
+                        if(!run_object(compl, nextin, nextout, runinfo)) {
+                            return FALSE;
+                        }
+                    }
+                    else {
+                        err_msg("There is no object", runinfo->mSName, runinfo->mSLine, runinfo->mArgsRuntime[i]);
+
+                        return FALSE;
+                    }
+                }
+
+                runinfo->mRCode = 0;
+            }
         }
     }
 
@@ -757,6 +813,26 @@ BOOL cmd_readline_clear_screen(sObject* nextin, sObject* nextout, sRunInfo* runi
     return TRUE;
 }
 
+BOOL cmd_readline_line_buffer(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
+{
+    if(!runinfo->mFilter) {
+        if(!fd_write(nextout, rl_line_buffer, strlen(rl_line_buffer))) {
+            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine, runinfo->mArgs[0]);
+            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
+            return FALSE;
+        }
+        if(!fd_write(nextout, "\n", 1)) {
+            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine, runinfo->mArgs[0]);
+            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
+            return FALSE;
+        }
+
+        runinfo->mRCode = 0;
+    }
+
+    return TRUE;
+}
+
 BOOL cmd_readline_point(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
 {
     if(!runinfo->mFilter) {
@@ -872,18 +948,6 @@ BOOL cmd_readline_write_history(sObject* nextin, sObject* nextout, sRunInfo* run
     return TRUE;
 }
 
-char* readline_filename_completion_null_generator(const char* a, int b)
-{
-    return NULL;
-}
-
-BOOL cmd_readline_file_name_completion_null_generator(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
-{
-    rl_completion_entry_function = readline_filename_completion_null_generator;
-
-    return TRUE;
-}
-
 static int readline_macro(int count, int key)
 {
     stack_start_stack();
@@ -1281,11 +1345,59 @@ static int lftp_char_is_quoted(const char *string, int eindex)
     }
   return (0);
 }
+void completion_display_matches_hook(char** matches, int num_matches, int max_length)
+{
+puts("");
+    int omit_head_of_string = 0;
+    sObject* readline = uobject_item(gRootObject, "rl");
+    if(readline && TYPE(readline) == T_UOBJECT) {
+        sObject* item = uobject_item(readline, "omit_head_of_completion_display_matches");
+        if(item && TYPE(item) == T_STRING) {
+            omit_head_of_string = atoi(string_c_str(item));
+        }
+    }
+
+    int my_max_length = 0;
+    int i;
+    for(i=1; i<=num_matches; i++) {
+        char* p = matches[i] + omit_head_of_string;
+        int length = str_termlen(kUtf8, p);
+        if(my_max_length < length) {
+            my_max_length = length;
+        }
+    }
+
+    my_max_length++;
+
+    const int maxx = mgetmaxx();
+    const int rows = maxx / my_max_length;
+
+    int x = 0;
+    for(i=1; i<=num_matches; i++) {
+        int x;
+        for(x=0; x<rows; x++) {
+            char* p = matches[i] + omit_head_of_string;
+            printf("%s", p);
+
+            int space_len = my_max_length - str_termlen(kUtf8, p);
+            int j;
+            for(j=0; j<space_len; j++) {
+                printf(" ");
+            }
+
+            i++;
+            if(i>num_matches) break;
+        }
+        puts("");
+    }
+    rl_forced_update_display();
+}
 
 void xyzsh_readline_init(BOOL runtime_script)
 {
     rl_attempted_completion_function = readline_on_complete;
-    //rl_completion_entry_function = readline_filename_completion_null_generator;
+    rl_completion_entry_function = readline_filename_completion_null_generator;
+
     rl_completer_quote_characters = "\"'";
     rl_completer_word_break_characters = " \t\n\"'|!&;()$<>=";
     rl_completion_append_character= ' ';
@@ -1294,9 +1406,30 @@ void xyzsh_readline_init(BOOL runtime_script)
     rl_filename_dequoting_function = (rl_dequote_func_t*)bash_dequote_filename;
     rl_char_is_quoted_p = (rl_linebuf_func_t*)lftp_char_is_quoted;
 
+    rl_completion_display_matches_hook = completion_display_matches_hook;
+
     rl_bind_key('x'-'a'+1, readline_macro);
 }
 
 void xyzsh_readline_final()
 {
 }
+
+void readline_object_init(sObject* self)
+{
+    sObject* readline = UOBJECT_NEW_GC(8, self, "rl", TRUE);
+    uobject_init(readline);
+    uobject_put(self, "rl", readline);
+
+    uobject_put(readline, "forced_update_display", NFUN_NEW_GC(cmd_readline_forced_update_display, NULL, TRUE));
+    uobject_put(readline, "insert_text", NFUN_NEW_GC(cmd_readline_insert_text, NULL, TRUE));
+    uobject_put(readline, "delete_text", NFUN_NEW_GC(cmd_readline_delete_text, NULL, TRUE));
+    uobject_put(readline, "clear_screen", NFUN_NEW_GC(cmd_readline_clear_screen, NULL, TRUE));
+    uobject_put(readline, "point_move", NFUN_NEW_GC(cmd_readline_point_move, NULL, TRUE));
+    uobject_put(readline, "read_history", NFUN_NEW_GC(cmd_readline_read_history, NULL, TRUE));
+    uobject_put(readline, "write_history", NFUN_NEW_GC(cmd_readline_write_history, NULL, TRUE));
+    uobject_put(readline, "replace_line", NFUN_NEW_GC(cmd_readline_replace_line, NULL, TRUE));
+    uobject_put(readline, "point", NFUN_NEW_GC(cmd_readline_point, NULL, TRUE));
+    uobject_put(readline, "line_buffer", NFUN_NEW_GC(cmd_readline_line_buffer, NULL, TRUE));
+}
+
