@@ -272,6 +272,7 @@ typedef struct {
 #define COMPLETION_FLAGS_STATMENT_END 0x8000
 #define COMPLETION_FLAGS_STATMENT_HEAD 0x10000
 #define COMPLETION_FLAGS_AFTER_REDIRECT 0x20000
+#define COMPLETION_FLAGS_AFTER_EQUAL 0x40000
 
 struct block_obj
 {
@@ -464,6 +465,7 @@ sObject* uobject_new_on_stack(int size, sObject* object, char* name);
 void uobject_put(sObject* self, char* key, void* item);
 void uobject_init(sObject* self);
 void uobject_root_init(sObject* self);
+void readline_object_init(sObject* self);
 void uobject_put(sObject* self, char* key, void* item);
 void* uobject_item(sObject* self, char* key);
 void uobject_clear(sObject* self);
@@ -563,20 +565,11 @@ typedef BOOL (*fInternalCommand)(sObject*, sObject*, sRunInfo*);
 extern fInternalCommand kInternalCommands[kInnerCommand];
 extern char * gCommandKindStrs[kCommandMax];
 
+BOOL cmd_readline(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_objinfo(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_kanjicode(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_defined(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_funinfo(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_forced_update_display(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_file_name_completion_null_generator(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_replace_line(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_point(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_read_history(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_write_history(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_point_move(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_clear_screen(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
-BOOL cmd_readline_insert_text(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_fselector(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_readline_delete_text(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
 BOOL cmd_mshow(sObject* nextin, sObject* nextout, sRunInfo* runinfo);
